import logging
from datetime import datetime
from typing import Dict, Union

from seastate.api.base import BaseApi
from seastate.api.rest_adapter import RestAdapter
from seastate.exceptions import SeaStateException
from seastate.models import Result


class TidesAndCurrentsApi(BaseApi):
    def __init__(self, logger: logging.Logger = None):
        """Constructor for TidesAndCurrentsApi, composed with RestAdapter"""
        self._logger = logger or logging.getLogger(__name__)
        self._adapter = RestAdapter("api.tidesandcurrents.noaa.gov")

    def _build_parse_key(self, measurement: str = None) -> Union[str, list[str], None]:
        return "data"

    def _build_endpoint(
        self, measurement: str, station_id: str, start: datetime, end: datetime
    ) -> (str, Dict):
        # each measurement type will have a different endpoint "product" param
        # always test str on lowercase casting
        measurement = measurement.lower()
        if "tide" in measurement:
            product = "water_level"
        elif "wind" in measurement:
            product = "wind"
        elif "water_temp" in measurement:
            product = "water_temperature"
        elif "air_temp" in measurement:
            product = "air_temperature"
        elif "air_press" in measurement:
            product = "air_pressure"
        elif "wave" in measurement:
            raise SeaStateException("Unsupported measurement requested")
        elif "conductivity" in measurement:
            product = "conductivity"
        else:
            raise SeaStateException("Unsupported measurement requested")

        # formatting datetime to endpoint param
        begin_date = f"{start.year}{start.month:02}{start.day:02}"
        end_date = f"{end.year}{end.month:02}{end.day:02}"

        ep_params = {
            "begin_date": begin_date,
            "end_date": end_date,
            "station": str(station_id),
            "product": product,
            "interval": "h",
            "datum": "MTL",
            "units": "metric",
            "time_zone": "lst_ldt",  # todo: local time or gmt? what is ndbc doing?
            "application": "seastate",
            "format": "json",
        }

        # Call endpoint
        endpoint = "api/prod/datagetter?"
        return endpoint, ep_params

    def _parse_result(
        self, result: list[Result], start: datetime, end: datetime, measurement: str
    ) -> list[Dict]:
        # unpack to return specified measurement
        # since TidesAndCurrents returns 1 product per endpoint:
        # -> minimal parsing, just unpack Json
        # details here: https://api.tidesandcurrents.noaa.gov/api/prod/responseHelp.html
        #
        if len(result) > 1:
            raise SeaStateException(
                "TidesAndCurrentsApi doesn't support multiple results"
            )
        try:
            parse_key = self._build_parse_key(measurement)
            # checking payload for error message first
            if result[0].data.get("error"):
                self._logger.error(
                    f"TidesAndCurrentsApi error, returning empty data\n{measurement}:{result[0].data}"
                )
                return []
            # parsing
            data = result[0].data[parse_key]
        except KeyError as e:
            msg = f"{e} unpacking {measurement} with key: {parse_key}\n{result[0].data}"
            self._logger.error(msg=msg)
            raise SeaStateException("TidesAndCurrentsApi unpacking error") from e

        if len(data) == 0:
            self._logger.warning(
                f"No {measurement} data recovered for daterange:\
                    {str(start.date())} : {str(end.date())}"
            )

        return data



import logging
from datetime import datetime, timedelta
from typing import Dict, Union

from seastate.api.api_mediator import ApiMediator
from seastate.settings import MEASUREMENTS


class SeaState:
    def __init__(
        self, lat: float, lon: float, exclude: list = [], logger: logging.Logger = None
    ):
        self._logger = logger or logging.getLogger(__name__)
        self.lat = float(lat)
        self.lon = float(lon)
        self.exclude = list(exclude)
        self.mediator_map = {}
        self._set_api_mediators()

    @property
    def _requested_measurements(self) -> tuple:
        return (x for x in MEASUREMENTS if x not in self.exclude)

    def _build_date_range(
        self, start: Union[datetime, None], end: Union[datetime, timedelta, None]
    ) -> tuple[datetime, datetime]:
        """
        Builds date range to fullest days possible,
        with start and end at midnight and (midnight - 1)
        """
        # Process timeframe
        # todo: handle iso date strings
        # todo: handle bad date strings
        if start and end:  # all values provided
            # end can be declared relative to start as a timedelta
            if isinstance(end, timedelta):
                end = start + end
        elif start and not end:  # no end provided, return same day as start
            end = start
        elif not start and not end:  # nothing provided, return today
            start = end = datetime.today()

        # remove microseconds to pass comparison tests
        # remove hours and minutes from start
        # set hours and minutes to end of day for end
        if isinstance(start, datetime):
            start = start.replace(microsecond=0)
            start = start.replace(hour=0, minute=0, second=0)
        if isinstance(end, datetime):
            end = end.replace(microsecond=0)
            end = end.replace(hour=23, minute=59, second=59)

        # log warning if end is before start
        if end < start:
            self._logger.warning("end is after start, swapping values")
            buffer = end
            end = start
            start = buffer

        return start, end

    def _get_mediator(self, measurement: str) -> ApiMediator:
        return self.__getattribute__(measurement)

    def _set_mediator(self, measurement: str) -> None:
        if measurement in self.exclude:
            self.__setattr__(measurement, None)
        else:
            tmp = ApiMediator(measurement, self.lat, self.lon, self.exclude)
            # add attribute
            self.__setattr__(measurement, tmp)
            # register mediator to map
            self.mediator_map[measurement] = tmp

    def _set_api_mediators(self) -> None:
        """set APIMediator as attribute for all implemented measurements"""
        for measurement in MEASUREMENTS:
            self._set_mediator(measurement)

    def _measurement_from_date_range(
        self, key: str, start: datetime, end: datetime
    ) -> list[Dict]:
        # selecting and configuring correct api mediator
        mediator = self._get_mediator(key)
        data = {}
        data = mediator.api._measurement_from_date_range(
            measurement=mediator.measurement,
            station_id=mediator.station.id,
            start=start,
            end=end,
        )
        return data

    def from_date_range(
        self,
        start: Union[datetime, None] = None,
        end: Union[datetime, timedelta, None] = None,
    ) -> Dict:
        # process timeframe
        start, end = self._build_date_range(start, end)

        # makes an api call for each measurement,
        # the idea is that a cache with 59 second expiration
        # absorbs the duplicated endpoint calls during a single method call
        data = {}
        for key in self._requested_measurements:
            try:
                data[key] = self._measurement_from_date_range(key, start, end)
            except Exception as e:
                self._logger.error(
                    f"Error occurred while retrieving data for measurement {key}: {str(e)}"
                )
                data[key] = []

        return data

    def hourly(
        self,
        start: Union[datetime, None] = None,
        end: Union[datetime, timedelta, None] = None,
    ) -> Dict:
        """Convenience method to return 1 sample per hour for each api"""
        data_all = self.from_date_range(start, end)
        data = {}
        for key in data_all:
            # find the first minute within the hour
            # the apis typically use the same minute
            first_minute = 0
            while datetime.fromisoformat(data_all[key][0]["t"]).minute != first_minute:
                first_minute += 1
                # prevent endless loop
                if first_minute == 60:
                    data[key] = []
                    break
            # keep one reading per hour
            data[key] = [
                x
                for x in data_all[key]
                if datetime.fromisoformat(x["t"]).minute == first_minute
            ]
        return data

    def measurements_from_date_range(
        self, start: datetime = None, end: Union[datetime, timedelta] = None
    ) -> Dict:
        self._logger.warning(
            "measurements_from_date_range() is deprecated,\
                use from_date_range() instead"
        )
        return self.from_date_range(start, end)


if __name__ == "__main__":
    api = SeaState(32, -117)
    start = datetime(2023, 9, 5)
    end = datetime(2023, 9, 5)
    data = api.from_date_range(start, end)
    a = api.hourly(start, end)

    # check daterange within realtime
    # result = api.hourly(datetime.today()-timedelta(days=2),datetime.today())

    # # check daterange for request only in prior years
    # result = api.hourly(
    #     datetime.today() - timedelta(days=2 * 365 + 30),
    #     datetime.today() - timedelta(days=1 * 365 + 30),
    # )

    # # check daterange spanning archive and realtime, into prior year
    # result = api.hourly(datetime.today()-timedelta(days=365+30),datetime.today())

    # # check old archive with different headers
    # # check date format
    # result = api.hourly('air_press',42040,datetime(1996,2,1),datetime(1996,2,2))



from typing import Union
from dataclasses import field
import logging

from seastate.data import STATIONS
from seastate.exceptions import SeaStateException
from seastate.models import Station
from seastate.utils import haversine
from seastate.api.base import BaseApi
from seastate.api.noaa_ndbc import NdbcApi
from seastate.api.noaa_tidesandcurrents import TidesAndCurrentsApi
from abc import ABC, abstractmethod


class BaseMediator(ABC):
    """Implements utils for finding nearest station and API for that station"""

    def __init__(self):
        raise NotImplementedError

    @abstractmethod
    def station(self) -> Station:
        pass

    @abstractmethod
    def api(self) -> BaseApi:
        pass

    @abstractmethod
    def distance(self) -> float:
        pass

    @property
    def _target_lat(self) -> None:
        return self.__target_lat

    @_target_lat.setter
    def _target_lat(self, value) -> None:
        #  Latitudes exist in range of -90 to 90
        if -90 <= value <= 90:
            self.__target_lat = value
        else:
            raise SeaStateException("Latitude must be between -90 and 90 degrees")

    @property
    def _target_lon(self) -> None:
        return self.__target_lon

    @_target_lon.setter
    def _target_lon(self, value) -> None:
        #  Longitudes exist in range of -180 to 180
        if -180 <= value <= 180:
            self.__target_lon = value
        else:
            raise SeaStateException("Longitude must be between -180 and 180 degrees")

    def _nearest_station(self) -> Station:
        """
        Find station that:
        - is closest to input coordinates using haversine method
        - is collecting the measurement at present (active)
        """
        station = None
        min_ptr = float("inf")  # Initialize minimum pointer at inf
        for eval_station in STATIONS:
            # skip if station is in exclude list
            if eval_station.id in self.exclude:
                continue
            # skip if eval_station is inactive
            if not eval_station.is_active:
                continue
            # skip if station does not support measurement
            if not eval_station.is_supported(self.measurement):
                continue
            # compute distance between SeaState target coords and current station
            new_val = haversine(
                self._target_lat, self._target_lon, eval_station.lat, eval_station.lon
            )
            # if closer, update new minimum
            if new_val < min_ptr:
                min_ptr = new_val
                station = eval_station
        return station


class ApiMediator(BaseMediator):
    """_summary_"""

    def __init__(
        self,
        measurement: str,
        lat: float,
        lon: float,
        exclude: list = field(default_factory=list),
        logger: logging.Logger = None,
    ):
        self._target_lat = lat
        self._target_lon = lon
        self.measurement = measurement
        self.exclude = exclude
        # self._register_mediator(self)
        self._logger = logger or logging.getLogger(__name__)

    @property
    def station(self) -> Station:
        return self._nearest_station()

    @property
    def api(self) -> Union[NdbcApi, TidesAndCurrentsApi]:
        if self.station.api == "noaa_ndbc":
            return NdbcApi()
        elif self.station.api == "noaa_tidesandcurrents":
            return TidesAndCurrentsApi()
        else:
            raise SeaStateException("Unsupported API")

    @property
    def distance(self) -> float:
        return haversine(
            self._target_lat, self._target_lon, self.station.lat, self.station.lon
        )


if __name__ == "__main__":
    api = ApiMediator("Tide", 32, -117)  # testing San Diego for tide
    # Station should have tide as a valid measurement
    assert api.station.tide
    assert api.station.is_supported(api.measurement)
    # out of bound lat/lon should throw errors

    # testing wind
    api = ApiMediator("Wind", 32, -117)  # testing San Diego for tide



from seastate.data import load_stations
from seastate.models import Station


def test_load_stations():
    stations = load_stations()
    assert len(stations) > 0
    assert isinstance(stations[0], Station)



from seastate.models import Station
from dataclasses import dataclass
from faker import Faker
from seastate.settings import DATASOURCES


@dataclass
class StationFactory:
    class Meta:
        model = Station

    id = Faker("id")
    lat = Faker().latitude()
    lon = Faker().longitude()
    api = Faker().random_element(elements=DATASOURCES)

    def create(self, **kwargs):
        return self.Meta.model(**kwargs)



from seastate.utils import haversine


def test_haversine_success():
    """Test haversine distance calculation"""
    assert haversine(32, -117, 32, -117) == 0
    assert haversine(32, -117, 33, -118) == 145.36863376095982



import logging
from typing import Dict, Union

import requests
import requests.packages
from requests_cache import CachedSession
from seastate.exceptions import SeaStateException
from seastate.models import Result


class RestAdapter:
    session = CachedSession(
        "seastate_cache", backend="filesystem", use_cache_dir=True, expire_after=59
    )

    def __init__(
        self,
        hostname: str,
        api_key: str = None,
        ssl_verify: bool = True,
        logger: logging.Logger = None,
    ):
        """Constructor for RestAdapter, supports GET

        Args:
            hostname (str): hostname for http request
            api_key (str, optional): Defaults to ''.
            ssl_verify (bool, optional): Defaults to True.
        """
        self._logger = logger or logging.getLogger(__name__)
        self.url = f"https://{hostname}/"
        self._api_key = api_key
        self._ssl_verify = ssl_verify
        if not ssl_verify:
            requests.packages.urllib3.disable_warnings()

    def _do(
        self, http_method: str, endpoint: str, ep_params: Dict = None, data: Dict = None
    ) -> Result:
        """HTTP request contructor convenience method

        Args:
            http_method (_type_): GET, POST, DELETE
            endpoint (str): target endpoint
            ep_params (Dict, optional): http parameters. Defaults to None.
            data (Dict, optional): POST data. Defaults to None.

        Returns:
            Result: Simplified http.Response object with:
                status_code
                message
                data
        """
        full_url = self.url + endpoint
        headers = {"x-api-key": self._api_key}
        log_line_pre = (
            f"request: method={http_method}, url={full_url}, params={ep_params}"
        )
        log_line_post = "result: success={}, status_code={}, message={}"
        # Log HTTP params and try HTTP request
        try:
            self._logger.debug(msg=log_line_pre)
            # response = requests.request(
            response = self.session.request(
                method=http_method,
                url=full_url,
                verify=self._ssl_verify,
                headers=headers,
                params=ep_params,
                json=data,
            )
        except Exception as e:
            self._logger.exception(msg=str(e))
            raise SeaStateException("Request Failed") from e
        # Parse JSON ouput to Python object or return failed Result on exception
        try:
            if ".txt" in full_url.lower():
                # handle txt based files from nonRest endpoints
                data_out = response.text
            elif "xml" in full_url.lower():
                # handle xml files from nonRest endpoints
                data_out = response.text
            elif any(x in full_url.lower() for x in ["spec", ".sw"]):
                # handle the 7 different spectral file extensions
                data_out = response.text
            else:
                # handle a restful endpoint
                data_out = response.json()
        except ValueError as e:
            self._logger.error(msg=log_line_post.format(False, None, e))
            raise SeaStateException("Bad JSON in Response") from e
        is_success = 299 >= response.status_code >= 200
        log_line = log_line_post.format(
            is_success, response.status_code, response.reason
        )
        # if status code in 200-299 range, return Result, else raise exception
        if is_success:
            self._logger.debug(msg=log_line)
            self._logger.debug(msg=log_line_pre)
            return Result(response.status_code, message=response.reason, data=data_out)
        # else, error has occured
        log_line += f"\n{log_line_pre}"  # enrich error msg
        self._logger.error(msg=log_line)
        raise SeaStateException(log_line)

    def get(self, endpoint: str, ep_params: Union[Dict, None] = None) -> Result:
        """GET method for RestAdapter

        Args:
            endpoint (str): target endpoint
            ep_params (Dict, optional): key:value API parameters. Defaults to None.

        Returns:
            Result: Simplified Response object with:
                status_code
                message
                data
        """
        return self._do(http_method="GET", endpoint=endpoint, ep_params=ep_params)



import glob
import os
from typing import List

from seastate.data.parsers import StationParser
from seastate.models import Station
from seastate.settings import DATASOURCES
from functools import lru_cache


def save_stations() -> None:
    """Calls implemented parsers and writes result to json"""
    parser = StationParser()
    basedir = os.path.dirname(__file__)

    for datasource in DATASOURCES:
        with open(os.path.join(basedir, f"{datasource}.json"), "w") as outfile:
            stations = getattr(parser, datasource)()
            data = parser.to_jsons(stations)
            outfile.write(data)


@lru_cache(maxsize=None)
def load_stations() -> List[Station]:
    """for all *.json in this directory, load into list of Station objects"""
    data = []
    for file in glob.glob(os.path.join(os.path.dirname(__file__), "*.json")):
        with open(file, "r") as content:
            data += StationParser.from_jsons(content.read())
    return data


if __name__ == "__main__":
    # write stations when called from CLI
    save_stations()
else:
    # load stations when imported
    STATIONS = load_stations()



from abc import ABC, abstractmethod
from seastate.settings import DATASOURCES
from seastate.exceptions import SeaStateException
from datetime import datetime
from typing import Dict, Optional, Union
from seastate.models import Result


class BaseApi(ABC):
    def __init__():
        raise NotImplementedError

    @property
    def id(self) -> str:
        """filename of the API module used internally as the id"""
        id = self.__module__.split(".")[-1]
        # integration check
        if id not in DATASOURCES:
            raise SeaStateException(f"API filename {id} does not match DATASOURCES")
        return id

    def _measurement_from_date_range(
        self, measurement: str, station_id: str, start: datetime, end: datetime
    ) -> list[Dict]:
        # build endpoint
        ep, ep_params = self._build_endpoint(measurement, station_id, start, end)
        # get result
        result = self._get_result(ep, ep_params)
        # parse result to data
        data = self._parse_result(result, start, end, measurement)
        return data

    @abstractmethod
    def _build_endpoint(
        self, measurement: str, station_id: str, start: datetime, end: datetime
    ) -> (str, Dict):
        pass

    @abstractmethod
    def _build_parse_key(self, measurement: str = None) -> Union[str, list[str], None]:
        pass

    def _get_result(
        self, endpoints: list[str], ep_params: Optional[Dict] = None
    ) -> list[Result]:
        """ep_params is applied to each endpoint in endpoints"""
        if isinstance(endpoints, str):
            endpoints = [endpoints]
        return [self._adapter.get(endpoint=ep, ep_params=ep_params) for ep in endpoints]

    @abstractmethod
    def _parse_result(
        self, result: list[Result], start: datetime, end: datetime, measurement: str
    ) -> Dict:
        pass



# the implemented measurements
MEASUREMENTS = (
    "tide",
    "wind",
    "water_temp",
    "air_temp",
    "air_press",
    "wave",
    "conductivity",
)

# the integrated APIs
# These are == to the filenames in api/ and /data
DATASOURCES = (
    "noaa_ndbc",
    "noaa_tidesandcurrents",
)



import json
import logging
from dataclasses import asdict
from typing import List

import defusedxml.minidom

from seastate.api.noaa_ndbc import NdbcApi
from seastate.api.noaa_tidesandcurrents import TidesAndCurrentsApi
from seastate.api.rest_adapter import RestAdapter
from seastate.exceptions import SeaStateException
from seastate.models import Station


class StationParser:
    def __init__(self, logger: logging.Logger = None):
        self._logger = logger or logging.getLogger(__name__)

    @staticmethod
    def from_jsons(jsons: str) -> List[Station]:
        """Parses a json string into a list of Station objects"""
        return [Station(**x) for x in json.loads(jsons)]

    @staticmethod
    def to_jsons(stations: List[Station]) -> str:
        """Parses a list of Station object into a json str"""
        return json.dumps([asdict(x) for x in stations])

    def noaa_ndbc(self) -> List[Station]:
        """
        Calls the ndbc endpoint
        summarizes all observations and parses the station metadata
        """
        # todo: refactor this spaghetti code
        # check station status by retrieving latest measurements from all stations
        result = RestAdapter("www.ndbc.noaa.gov/").get("data/latest_obs/latest_obs.txt")

        # loop through text lines and test if station is ouputting specified measurement
        stations = []
        for i, value in enumerate(result.data.split("\n")):
            # skipping 2 header lines
            if i == 0 or i == 1:
                continue  # skip header and unit lines

            # split current line by delimiter and clear empty values
            line = value.split(" ")
            while " " in line:
                line.remove(" ")
            while "" in line:
                line.remove("")

            # skip corrupted lines
            if len(line) != 22:
                self._logger.info(f"No station info in this line: {str(line)}")
                continue

            # parse stationID, gps, and confirm active measurement sources
            # 'MM' is NOAA's notation for missing measurement
            tmp_station = {
                "id": line[0],
                "lat": float(line[1]),
                "lon": float(line[2]),
                "api": NdbcApi().id,
                "tide": True if line[21] != "MM" else False,
                "wind": True if line[8] != "MM" else False,
                "water_temp": True if line[18] != "MM" else False,
                "air_temp": True if line[17] != "MM" else False,
                "air_press": True if line[15] != "MM" else False,
                "wave": True if line[11] != "MM" else False,
            }

            # construct station with temp var
            stations.append(Station(**tmp_station))

        # return list of stations
        return stations

    def noaa_tidesandcurrents(self) -> List[Station]:
        """
        Calls the tnc endpoint
        summarizes all stations and parses their metadata
        """
        # todo: refactor this spaghetti code
        # TidesAndCurrent station capabilities are kept here
        result = RestAdapter("opendap.co-ops.nos.noaa.gov/").get(
            "stations/stationsXML.jsp"
        )

        # parse xml elements safely with defusedxml -> []
        dom = defusedxml.minidom.parseString(result.data)
        station_elements = dom.getElementsByTagName("station")

        # traverse station elements to build Station objects
        stations = []
        for line in station_elements:
            tmp_station = {}
            # parse station metadata into temporary dict
            try:
                tmp_station["name"] = line.getAttribute("name")
                tmp_station["id"] = line.getAttribute("ID")
                tmp_station["lat"] = float(
                    line.getElementsByTagName("lat")[0].firstChild.nodeValue
                )
                tmp_station["lon"] = float(
                    line.getElementsByTagName("long")[0].firstChild.nodeValue
                )
                tmp_station["api"] = TidesAndCurrentsApi().id
            except IndexError as e:
                # Faulty tides and currents station, skip station row
                self._logger.warn(f"{e}: {str(line)}")
                continue

            # separate loop to parse individual measurements
            for m in line.getElementsByTagName("parameter"):
                name = m.attributes["name"].value
                status = True if m.attributes["status"].value == "1" else False
                if "Water Level" in name and status:
                    tmp_station["tide"] = True
                elif "Winds" in name and status:
                    tmp_station["wind"] = True
                    # todo: wind dir and gust may or may not be true
                elif "Air Temp" in name and status:
                    tmp_station["air_temp"] = True
                elif "Water Temp" in name and status:
                    tmp_station["water_temp"] = True
                elif "Air Pressure" in name and status:
                    tmp_station["air_press"] = True
                elif "Conductivity" in name and status:
                    tmp_station["conductivity"] = True

            # construct station with temp var
            stations.append(Station(**tmp_station))

        # Check parsing was succesful
        if len(stations) == 0:
            raise SeaStateException(
                "No stations successfully parsed, please submit issue"
            )
        return stations



from seastate import SeaState
from datetime import datetime, timedelta
import pytest
from seastate.settings import MEASUREMENTS


class TestSeaState:
    @pytest.fixture
    def seastate(self):
        return SeaState(32, -117)

    @pytest.fixture
    def seastate_exclude(self):
        return SeaState(32, -117, exclude=["tide", "13009", "noaa_ndbc"])

    # Tests
    def test_set_mediators_success(self, seastate_exclude):
        # all measurements should be set as attributes
        # -> intersection of measurements and __dict__ should be length of measurements
        keys = set([x for x in seastate_exclude.__dict__.keys()])
        intersection = [x for x in keys if x in MEASUREMENTS]
        assert len(MEASUREMENTS) == len(intersection)

    def test_exclude_measurement_success(self, seastate_exclude):
        # assert seastate does not have tide
        assert seastate_exclude.tide is None

    @pytest.mark.skip
    def test_exclude_station_success(self, seastate_exclude):
        assert False

    @pytest.mark.skip
    def test_exclude_api_success(self, seastate_exclude):
        assert False

    start = datetime(2023, 10, 5)
    end = datetime(2023, 10, 6)
    delta = timedelta(days=1)

    @pytest.mark.parametrize(
        "start, end, expected",
        [
            (start, end, 2 * 24 * 60 * 60 - 1),
            (start, delta, 2 * 24 * 60 * 60 - 1),
            (start, None, 1 * 24 * 60 * 60 - 1),
            (None, None, 1 * 24 * 60 * 60 - 1),
        ],
    )
    def test_build_date_range_success(self, seastate, start, end, expected):
        # simplistic test,
        # should return a daterange with width 2 days or 1 day, minus 1 second
        start, end = seastate._build_date_range(start, end)
        assert (end - start).total_seconds() == expected



from seastate.data.parsers import StationParser
import json
from tests.factories import StationFactory


class TestStationParser:
    TEST_LIST = [
        {"id": "Station A", "lat": 51.5074, "lon": -0.1278, "api": "noaa_ndbc"},
        {"id": "Station B", "lat": 52.5074, "lon": -1.1278, "api": "noaa_ndbc"},
    ]

    TEST_STRING = '[{"id": "Station A", \
        "lat": 51.5074, "lon": -0.1278, "api": "noaa_ndbc"},\
        {"id": "Station B", \
            "lat": 52.5074, "lon": -1.1278, "api": "noaa_ndbc"}]'

    def test_to_jsons_shallow_success(self):
        parser = StationParser()
        stations = [StationFactory().create(**x) for x in self.TEST_LIST]
        jsons = parser.to_jsons(stations)
        # just checking that length is the same, not full identity
        assert len(json.loads(jsons)) == len(self.TEST_LIST)

    def test_from_jsons_shallow_success(self):
        parser = StationParser()
        jsons = self.TEST_STRING
        stations = parser.from_jsons(jsons)
        # just checking that length is the same, not full identity
        assert len(stations) == 2



from seastate.api.api_mediator import ApiMediator
import pytest


class TestApiMediator:
    @pytest.fixture
    def mediator(self):
        return ApiMediator("wind", 32, -117)

    @pytest.mark.skip
    def test_1(self):
        # init
        assert False



import logging
from datetime import datetime, timedelta
from typing import Dict, Union

from seastate.api.base import BaseApi
from seastate.api.rest_adapter import RestAdapter
from seastate.exceptions import SeaStateException
from seastate.models import Result


class NdbcApi(BaseApi):
    def __init__(self, logger: logging.Logger = None):
        """Constructor for NdbcApi, composed with RestAdapter"""
        self._logger = logger or logging.getLogger(__name__)
        self._adapter = RestAdapter("www.ndbc.noaa.gov")

    def _build_parse_key(self, measurement: str = None) -> Union[str, list[str], None]:
        measurement = measurement.lower()
        if "tide" in measurement:
            key = ["TIDE"]
        elif "wind" in measurement:
            key = ["WSPD", "SPD"]
            # additional wind information handled when unpacking
        elif "water_temp" in measurement:
            key = ["WTMP"]
        elif "air_temp" in measurement:
            key = ["ATMP"]
        elif "air_press" in measurement:
            key = ["PRES", "BARO", "BAR"]
        elif "wave" in measurement:
            key = ["WVHT", "H0"]
            # additional swell information handled when unpacking
        elif "conductivity" in measurement:
            raise SeaStateException(
                "Unsupported measurement requested, please report issue"
            )
        else:
            raise SeaStateException(
                "Unsupported measurement requested, please report issue"
            )
        return key

    def _build_endpoint(
        self, measurement: str, station_id: str, start: datetime, end: datetime
    ) -> (str, Dict):
        # Collect endpoints needed to cover daterange
        endpoints = []
        # daterange spans realtime endpoint (45 days)
        archive_cutoff_date = datetime.today() - timedelta(days=45)
        if start > archive_cutoff_date or end > archive_cutoff_date:
            # "realtime" endpoint covers prev 45 days
            endpoints += [f"data/realtime2/{station_id}.txt"]
        # daterange spans archive, for current year
        # cutoff month has a unique endpoint
        if start < archive_cutoff_date < end:
            requested_months = [archive_cutoff_date.strftime("%b")]
            endpoints += [f"data/stdmet/{x}/{station_id}.txt" for x in requested_months]
        # within current year, endpoints are grouped by month
        if start < archive_cutoff_date and end.year == datetime.today().year:
            # use January if start year precedes the archive cutoff year
            # else use start month
            start_range = 1 if start.year < archive_cutoff_date.year else start.month
            # end of range will be up to cutoff
            # duplicates to be filtered downstream
            end_range = archive_cutoff_date.month
            # Endpoint uses month number and abbreviated month designations (%b)
            requested_months = [
                (str(x), datetime(1, x, 1).strftime("%b"))
                for x in range(start_range, end_range)
            ]
            endpoints += [
                f"view_text_file.php?filename= \
                    {station_id}{x}{datetime.today().year} \
                        .txt.gz&dir=data/stdmet/{y}/"
                for x, y in requested_months
            ]
        # daterange spans prior years
        #
        if start < archive_cutoff_date and start.year < datetime.today().year:
            # if end year is this year, handled upstream
            # else, it is a prior year and needs to be included
            end_range = end.year if end.year == datetime.today().year else end.year + 1
            requested_years = [str(x) for x in range(start.year, end_range)]
            endpoints += [
                f"view_text_file.php?filename= \
                    {station_id}h{x} \
                        .txt.gz&dir=data/historical/stdmet/"
                for x in requested_years
            ]
        return endpoints, None

    def _parse_result(
        self, result: list[Result], start: datetime, end: datetime, measurement: str
    ) -> list[Dict]:
        # unpack list of results to text
        tmp = ""
        for res in result:
            tmp += res.data

        # unpack text result
        # realtime ndbc reports are a text file with previous 45 days of measurements
        # archival are a years worth
        # todo: unpack text file into t: and v:
        data = []
        for i, value in enumerate(tmp.split("\n")):
            # turn str to list
            # known anomaly: 2 space characters sometimes delimiting
            line = value.replace("  ", " ").split(" ")
            while " " in line:
                line.remove(" ")  # remove blanks
            while "" in line:
                line.remove("")  # remove blanks

            # handle corrupted lines
            if len(line) == 0:
                self._logger.info(f"ignoring line #{i}:{str(line)}")
                continue

            # handle header lines
            if line[0] in ["#YY", "YY"]:
                # header row starts with #YY in realtime, YY in archive
                header = line
                # strip '#' from '#YY'
                header[0] = header[0].strip("#")
                continue
            elif line[0] in ["#yr"]:
                # skipping units line
                continue

            # ignore lines outside of daterange
            # archive files sometimes use 95 instead of 1995
            if not (start.year <= int(line[0]) <= end.year) and not (
                abs(start.year % 100) <= int(line[0]) <= abs(end.year % 100)
            ):
                # 4 digit year representation
                continue
            if not start.month <= int(line[1]) <= end.month:
                continue
            if not start.day <= int(line[2]) <= end.day:
                continue

            # line is in requested daterange
            # parse timestamp
            timestamp = datetime(
                year=int(
                    line[0] if len(line[0]) == 4 else "19" + line[0]
                ),  # rebuild dates with archive format
                month=int(line[1]),
                day=int(line[2]),
                hour=int(line[3]),
                minute=int(
                    line[4] if "mm" in header else 0
                ),  # archival format is hourly
            ).isoformat(sep=" ")
            # parse measurement column
            # because of changes in the source api column names over the years
            # we try to unpack with the possible variants
            # details here: https://www.ndbc.noaa.gov/measdes.shtml
            temp_data = {}
            # unpack time and main value
            for key in self._build_parse_key(measurement):
                try:
                    temp_data["t"] = timestamp
                    temp_data["v"] = line[header.index(key)]
                except (KeyError, ValueError):
                    pass
            # check for success before continuing
            if len(temp_data) < 2:
                self._logger.error(value)
                raise SeaStateException("NdbcApi unpacking error, please report issue")
            # wind sometimes has direction and gust data
            if "wind" in measurement:
                # for wind direction:
                for key in ["WDIR", "WD", "DIR"]:
                    try:
                        temp = line[header.index(key)]
                        temp_data["d"] = (
                            temp
                            if temp and "999" not in temp and "MM" not in temp
                            else None
                        )  # 999 for direction
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
                # For wind gust:
                for key in ["GST", "GSP"]:
                    try:
                        temp = line[header.index(key)]
                        temp_data["g"] = (
                            temp
                            if temp and "99" not in temp and "MM" not in temp
                            else None
                        )  # 99 for decimal
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
            # wave sometimes has period and direction
            if "wave" in measurement:
                # for dominant wave period:
                for key in ["DPD", "DOMPD"]:
                    try:
                        temp = line[header.index(key)]
                        temp_data["dpd"] = (
                            temp
                            if temp and "99" not in temp and "MM" not in temp
                            else None
                        )  # 99 for decimal
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
                # For dominant wave direction:
                for key in ["MWD"]:
                    try:
                        temp = line[header.index(key)]
                        temp_data["mwd"] = (
                            temp
                            if temp and "999" not in temp and "MM" not in temp
                            else None
                        )  # 999 for direction
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
                # For average wave period
                for key in ["APD", "AVP"]:
                    try:
                        temp = line[header.index(key)]
                        temp_data["apd"] = (
                            temp
                            if temp and "99" not in temp and "MM" not in temp
                            else None
                        )
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
            data.append(temp_data)

        # log warning if no data recovered for daterange
        if len(data) == 0:
            self._logger.warning(
                f"No {measurement} data recovered for daterange:\
                    {str(start.date())} : {str(end.date())}"
            )

        # todo: scrub duplicates between cutoff month and realtime
        return data


if __name__ == "__main__":
    api = NdbcApi()
    # check daterange within realtime
    # result = api.hourly(
    #     "wind",
    #     46224,
    #     datetime.today(),
    #     datetime.today() - timedelta(days=2),
    # )

    # check daterange for request only in prior years
    # result = api.hourly(
    #     "wind",
    #     46224,
    #     datetime.today() - timedelta(days=2 * 365 + 30),
    #     datetime.today() - timedelta(days=1 * 365 + 30),
    # )

    # check daterange spanning archive and realtime, into prior year
    # result = api.hourly(
    #     "wind",
    #     46224,
    #     datetime.today() - timedelta(days=365 + 30),
    #     datetime.today(),
    # )

    # check old archive with different headers
    # check date format
    result = api.hourly(
        "air_press",
        42040,
        datetime(1996, 2, 1),
        datetime(1996, 2, 2),
    )



from seastate.api.noaa_ndbc import NdbcApi


class TestNdbcApi:
    def test_id(self):
        assert NdbcApi().id == "noaa_ndbc"


# from seastate.models import Result
# from seastate.exceptions import SeaStateException
# import requests

# def TestTideApi(TestCase):
#     def setUp(self):
#         self._tide_api = TideApi()
#         self.response = requests.Response()

#     # hourly tests
#     def test_good_call_returns_result(self):
#         assert False

#     def test_good_df_call_returns_df(self):
#         assert False

#     def test_empty_call_raises_error(self):
#         assert False

#     # measurement switch
#     def test_bad_measurement_raises_keyerror(self):
#         assert False

#     # time window feature
#     def test_no_timestring_raises_valueerror(self):
#         assert False

#     def test_bad_timestring_raises_typeerror(self):
#         assert False

#     def test_historic_timeframe_is_handled(self):
#         assert False

#     # time offset feature
#     def test_good_call_with_timeoffset_returns_result(self):
#         assert False

#     def test_bad_timeoffset_raises_typeerror(self):
#         assert False



from seastate.api.rest_adapter import RestAdapter
from seastate.models import Result
from seastate.exceptions import SeaStateException

import pytest


# needs fixture for adapter
class TestRestAdapter:
    @pytest.fixture
    def adapter(self):
        return RestAdapter("google.com")

    # _do
    def test__do_good_request_returns_result(self, adapter):
        result = adapter._do("GET", "robots.txt")
        assert isinstance(result, Result)

    @pytest.mark.skip
    def test__do_bad_json_raises_SeaStateException(self, adapter):
        result = adapter._do("GET", "")  # returning htnl counts as bad json
        assert isinstance(result, SeaStateException)



from seastate import SeaState
from datetime import datetime, timedelta
import pytest
from seastate.settings import MEASUREMENTS

# todo: add reg test for historical endpoints
# todo: cover ndbc and tnc datasources
# todo: write unit test for when result = emtpy, return [] and logger warning, for both datasources


class TestSeaState:
    @pytest.fixture
    def seastate(self):
        return SeaState(35.81468, -122.78828)

    # timestamps
    start = datetime(2023, 10, 5)
    end = datetime(2023, 10, 6)
    delta = timedelta(days=1)
    today = datetime.now()
    yesterday = today - timedelta(days=1)

    @pytest.mark.parametrize(
        "start, end",
        [
            (today, today),
            (yesterday, yesterday),
            # (start, end),
        ],
    )
    def test_from_date_range_success(self, seastate, start, end):
        # should return a dict with keys matching MEASUREMENTS
        data = seastate.from_date_range(start, end)
        assert set(data.keys()) == set(MEASUREMENTS)


# Prev test cases
# https://www.ndbc.noaa.gov/data/realtime2/42040.txt
# https://www.ndbc.noaa.gov/data/stdmet/May/42040.txt
# https://www.ndbc.noaa.gov/view_text_file.php?filename=42040h1996.txt.gz&dir=data/historical/stdmet/
# https://www.ndbc.noaa.gov/view_text_file.php?filename=42040h1999.txt.gz&dir=data/historical/stdmet/
# https://www.ndbc.noaa.gov/data/realtime2/1801589.txt



import numpy as np


def haversine(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """Returns distance in km using haversine method

    Args:
        lat1 (float): Coordinate in decimal degrees
        lon1 (float): Coordinate in decimal degrees
        lat2 (float): Coordinate in decimal degrees
        lon2 (float): Coordinate in decimal degrees

    Returns:
        float: Returns great circle distance in km
    """
    # Convert degrees to radians
    lat1, lon1, lat2, lon2 = map(np.radians, [lat1, lon1, lat2, lon2])

    # Calculate deltas
    dlat = lat2 - lat1
    dlon = lon2 - lon1

    # Haversine method for measuring distance on a sphere
    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2
    c = 2 * np.arcsin(np.sqrt(a))

    km = 6367 * c
    return km
